//////////////////////////////////////////////////////////////////////////////////////////////
// Copyright 2018 Leszek Koltunski                                                          //
//                                                                                          //
// This file is part of Distorted.                                                          //
//                                                                                          //
// Distorted is free software: you can redistribute it and/or modify                        //
// it under the terms of the GNU General Public License as published by                     //
// the Free Software Foundation, either version 2 of the License, or                        //
// (at your option) any later version.                                                      //
//                                                                                          //
// Distorted is distributed in the hope that it will be useful,                             //
// but WITHOUT ANY WARRANTY; without even the implied warranty of                           //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            //
// GNU General Public License for more details.                                             //
//                                                                                          //
// You should have received a copy of the GNU General Public License                        //
// along with Distorted.  If not, see <http://www.gnu.org/licenses/>.                       //
//////////////////////////////////////////////////////////////////////////////////////////////

precision highp float;
precision highp int;

out vec4 fragColor;           // The output color
in vec2 v_TexCoordinate;      // Interpolated texture coordinate per fragment.
in vec2 v_Pixel;              // location of the current fragment, in pixels

uniform sampler2D u_Texture;
uniform sampler2D u_DepthTexture;

//////////////////////////////////////////////////////////////////////////////////////////////
// per-pixel linked list. Order Independent Transparency.

uniform vec2 u_Size;

layout (std430,binding=1) buffer linkedlist  // first (u_Size.x*u_Size.y) uints - head pointers,
  {                                          // one for each pixel in the Output rectangle.
  uint u_Records[];                          //
  };                                         // Next 3*u_numRecords uints - actual linked list, i.e.
                                             // triplets of (pointer,depth,rgba).

//////////////////////////////////////////////////////////////////////////////////////////////

vec4 convert(uint rgba)
  {
  return vec4( float((rgba>>24u)&255u),float((rgba>>16u)&255u),float((rgba>>8u)&255u),float(rgba&255u) ) / 255.0;
  }

//////////////////////////////////////////////////////////////////////////////////////////////
// https://en.wikipedia.org/wiki/Alpha_compositing (premultiplied)

vec4 blend(vec4 clr,vec4 srf)
  {
  return clr + (1.0 - clr.a) * vec4(srf.rgb * srf.a , srf.a);
  }

//////////////////////////////////////////////////////////////////////////////////////////////
// Pass3 of the OIT algorithm - traverse the per-pixel LinkedList and build the final color.

void main()                    		
  {
  float texdepth = texture(u_DepthTexture, v_TexCoordinate).r;
  vec4  color    = texture(u_Texture     , v_TexCoordinate);
  uint  index    = uint(v_Pixel.x + v_Pixel.y * u_Size.x);
  uint  curr     = u_Records[index];

  if (curr != 0u)
    {
    const float S= 2147483647.0;
    uint depth = u_Records[curr+1u];
    uint texdepthuint = uint(S*(1.0-texdepth)/2.0);
    texdepth = 1.0 -2.0*float(depth)/S;

    if( depth >= texdepthuint )
      {
      vec4 clr= convert(u_Records[curr+2u]);
      curr = u_Records[curr];

      while (curr > 0u)
        {
        depth= u_Records[curr+1u];                       // keep walking the linked list
        if( depth < texdepthuint ) break;                // until we reach scene depth
        clr= blend( clr, convert(u_Records[curr+2u]) );  // and blending the colors in
        curr = u_Records[curr];
        }

      color = blend( clr, color);
      }
    }

  gl_FragDepth = texdepth;
  fragColor    = color;
  }